/*
 * Copyright (C) 1994-2021 Altair Engineering, Inc.
 * For more information, contact Altair at www.altair.com.
 *
 * This file is part of both the OpenPBS software ("OpenPBS")
 * and the PBS Professional ("PBS Pro") software.
 *
 * Open Source License Information:
 *
 * OpenPBS is free software. You can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 *
 * OpenPBS is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Commercial License Information:
 *
 * PBS Pro is commercially licensed software that shares a common core with
 * the OpenPBS software.  For a copy of the commercial license terms and
 * conditions, go to: (http://www.pbspro.com/agreement.html) or contact the
 * Altair Legal Department.
 *
 * Altair's dual-license business model allows companies, individuals, and
 * organizations to create proprietary derivative works of OpenPBS and
 * distribute them - whether embedded or bundled with other software -
 * under a commercial license agreement.
 *
 * Use of Altair's trademarks, including but not limited to "PBS™",
 * "OpenPBS®", "PBS Professional®", and "PBS Pro™" and Altair's logos is
 * subject to Altair's trademark licensing policies.
 */

/**
 *
 * @brief
 *      Implementation of the resv data access functions for postgres
 */
#include <pbs_config.h> /* the master config generated by configure */
#include "pbs_db.h"
#include "db_postgres.h"

/**
 * @brief
 *	Prepare all the reservation related sqls. Typically called after connect
 *	and before any other sql exeuction
 *
 * @param[in]	conn - Database connection handle
 *
 * @return      Error code
 * @retval	-1 - Failure
 * @retval	 0 - Success
 *
 */
int
db_prepare_resv_sqls(void *conn)
{
	char conn_sql[MAX_SQL_LENGTH];
	snprintf(conn_sql, MAX_SQL_LENGTH, "insert into pbs.resv ("
					   "ri_resvID, "
					   "ri_queue, "
					   "ri_state, "
					   "ri_substate, "
					   "ri_stime, "
					   "ri_etime, "
					   "ri_duration, "
					   "ri_tactive, "
					   "ri_svrflags, "
					   "ri_savetm, "
					   "ri_creattm, "
					   "attributes "
					   ") "
					   "values "
					   "($1, $2, $3, $4, $5, $6, $7, $8, $9, "
					   "localtimestamp, localtimestamp, hstore($10::text[]))");
	if (db_prepare_stmt(conn, STMT_INSERT_RESV, conn_sql, 10) != 0)
		return -1;

	snprintf(conn_sql, MAX_SQL_LENGTH, "update pbs.resv set "
					   "ri_queue = $2, "
					   "ri_state = $3, "
					   "ri_substate = $4, "
					   "ri_stime = $5, "
					   "ri_etime = $6, "
					   "ri_duration = $7, "
					   "ri_tactive = $8, "
					   "ri_svrflags = $9, "
					   "ri_savetm = localtimestamp, "
					   "attributes = attributes || hstore($10::text[]) "
					   "where ri_resvID = $1");
	if (db_prepare_stmt(conn, STMT_UPDATE_RESV, conn_sql, 10) != 0)
		return -1;

	snprintf(conn_sql, MAX_SQL_LENGTH, "update pbs.resv set "
					   "ri_queue = $2, "
					   "ri_state = $3, "
					   "ri_substate = $4, "
					   "ri_stime = $5, "
					   "ri_etime = $6, "
					   "ri_duration = $7, "
					   "ri_tactive = $8, "
					   "ri_svrflags = $9, "
					   "ri_savetm = localtimestamp "
					   "where ri_resvID = $1");
	if (db_prepare_stmt(conn, STMT_UPDATE_RESV_QUICK, conn_sql, 9) != 0)
		return -1;

	snprintf(conn_sql, MAX_SQL_LENGTH, "update pbs.resv set "
					   "ri_savetm = localtimestamp, "
					   "attributes = attributes || hstore($2::text[]) "
					   "where ri_resvID = $1");
	if (db_prepare_stmt(conn, STMT_UPDATE_RESV_ATTRSONLY, conn_sql, 2) != 0)
		return -1;

	snprintf(conn_sql, MAX_SQL_LENGTH, "update pbs.resv set "
					   "ri_savetm = localtimestamp,"
					   "attributes = attributes - $2::text[] "
					   "where ri_resvID = $1");
	if (db_prepare_stmt(conn, STMT_REMOVE_RESVATTRS, conn_sql, 2) != 0)
		return -1;

	snprintf(conn_sql, MAX_SQL_LENGTH, "select "
					   "ri_resvID, "
					   "ri_queue, "
					   "ri_state, "
					   "ri_substate, "
					   "ri_stime, "
					   "ri_etime, "
					   "ri_duration, "
					   "ri_tactive, "
					   "ri_svrflags, "
					   "hstore_to_array(attributes) as attributes "
					   "from pbs.resv where ri_resvid = $1");
	if (db_prepare_stmt(conn, STMT_SELECT_RESV, conn_sql, 1) != 0)
		return -1;

	snprintf(conn_sql, MAX_SQL_LENGTH, "select "
					   "ri_resvID, "
					   "ri_queue, "
					   "ri_state, "
					   "ri_substate, "
					   "ri_stime, "
					   "ri_etime, "
					   "ri_duration, "
					   "ri_tactive, "
					   "ri_svrflags, "
					   "hstore_to_array(attributes) as attributes "
					   "from pbs.resv "
					   "order by ri_creattm");
	if (db_prepare_stmt(conn, STMT_FINDRESVS_ORDBY_CREATTM,
			    conn_sql, 0) != 0)
		return -1;

	snprintf(conn_sql, MAX_SQL_LENGTH, "delete from pbs.resv where ri_resvid = $1");
	if (db_prepare_stmt(conn, STMT_DELETE_RESV, conn_sql, 1) != 0)
		return -1;

	return 0;
}

/**
 * @brief
 *	Load resv data from the row into the resv object
 *
 * @param[in]	res - Resultset from a earlier query
 * @param[in]	presv  - resv object to load data into
 * @param[in]	row - The current row to load within the resultset
 *
 * @return      Error code
 * @retval	-1 - On Error
 * @retval	 0 - On Success
 * @retval	>1 - Number of attributes
 */
static int
load_resv(PGresult *res, pbs_db_resv_info_t *presv, int row)
{
	char *raw_array;
	static int ri_resvid_fnum;
	static int ri_queue_fnum;
	static int ri_state_fnum;
	static int ri_substate_fnum;
	static int ri_stime_fnum;
	static int ri_etime_fnum;
	static int ri_duration_fnum;
	static int ri_tactive_fnum;
	static int ri_svrflags_fnum;
	static int attributes_fnum;
	static int fnums_inited = 0;

	if (fnums_inited == 0) {
		ri_resvid_fnum = PQfnumber(res, "ri_resvID");
		ri_queue_fnum = PQfnumber(res, "ri_queue");
		ri_state_fnum = PQfnumber(res, "ri_state");
		ri_substate_fnum = PQfnumber(res, "ri_substate");
		ri_stime_fnum = PQfnumber(res, "ri_stime");
		ri_etime_fnum = PQfnumber(res, "ri_etime");
		ri_duration_fnum = PQfnumber(res, "ri_duration");
		ri_tactive_fnum = PQfnumber(res, "ri_tactive");
		ri_svrflags_fnum = PQfnumber(res, "ri_svrflags");
		attributes_fnum = PQfnumber(res, "attributes");
		fnums_inited = 1;
	}

	GET_PARAM_STR(res, row, presv->ri_resvid, ri_resvid_fnum);
	GET_PARAM_STR(res, row, presv->ri_queue, ri_queue_fnum);
	GET_PARAM_INTEGER(res, row, presv->ri_state, ri_state_fnum);
	GET_PARAM_INTEGER(res, row, presv->ri_substate, ri_substate_fnum);
	GET_PARAM_BIGINT(res, row, presv->ri_stime, ri_stime_fnum);
	GET_PARAM_BIGINT(res, row, presv->ri_etime, ri_etime_fnum);
	GET_PARAM_BIGINT(res, row, presv->ri_duration, ri_duration_fnum);
	GET_PARAM_INTEGER(res, row, presv->ri_tactive, ri_tactive_fnum);
	GET_PARAM_INTEGER(res, row, presv->ri_svrflags, ri_svrflags_fnum);
	GET_PARAM_BIN(res, row, raw_array, attributes_fnum);

	/* convert attributes from postgres raw array format */
	return (dbarray_to_attrlist(raw_array, &presv->db_attr_list));
}

/**
 * @brief
 *	Insert resv data into the database
 *
 * @param[in]	conn - Connection handle
 * @param[in]	obj  - Information of resv to be inserted
 *
 * @return      Error code
 * @retval	-1 - Failure
 * @retval	 0 - Success
 *
 */
int
pbs_db_save_resv(void *conn, pbs_db_obj_info_t *obj, int savetype)
{
	pbs_db_resv_info_t *presv = obj->pbs_db_un.pbs_db_resv;
	char *stmt = NULL;
	int params;
	int rc = 0;
	char *raw_array = NULL;

	SET_PARAM_STR(conn_data, presv->ri_resvid, 0);

	if (savetype & OBJ_SAVE_QS) {
		SET_PARAM_STR(conn_data, presv->ri_queue, 1);
		SET_PARAM_INTEGER(conn_data, presv->ri_state, 2);
		SET_PARAM_INTEGER(conn_data, presv->ri_substate, 3);
		SET_PARAM_BIGINT(conn_data, presv->ri_stime, 4);
		SET_PARAM_BIGINT(conn_data, presv->ri_etime, 5);
		SET_PARAM_BIGINT(conn_data, presv->ri_duration, 6);
		SET_PARAM_INTEGER(conn_data, presv->ri_tactive, 7);
		SET_PARAM_INTEGER(conn_data, presv->ri_svrflags, 8);
		stmt = STMT_UPDATE_RESV_QUICK;
		params = 9;
	}

	if ((presv->db_attr_list.attr_count > 0) || (savetype & OBJ_SAVE_NEW)) {
		int len = 0;
		/* convert attributes to postgres raw array format */
		if ((len = attrlist_to_dbarray(&raw_array, &presv->db_attr_list)) <= 0)
			return -1;

		if (savetype & OBJ_SAVE_QS) {
			SET_PARAM_BIN(conn_data, raw_array, len, 9);
			stmt = STMT_UPDATE_RESV;
			params = 10;
		} else {
			SET_PARAM_BIN(conn_data, raw_array, len, 1);
			params = 2;
			stmt = STMT_UPDATE_RESV_ATTRSONLY;
		}
	}

	if (savetype & OBJ_SAVE_NEW)
		stmt = STMT_INSERT_RESV;

	if (stmt)
		rc = db_cmd(conn, stmt, params);

	return rc;
}

/**
 * @brief
 *	Load resv data from the database
 *
 * @param[in]	conn - Connection handle
 * @param[in]	obj  - Load resv information into this object
 *
 * @return      Error code
 * @retval	-1 - Failure
 * @retval	 0 - Success
 * @retval	 1 -  Success but no rows loaded
 *
 */
int
pbs_db_load_resv(void *conn, pbs_db_obj_info_t *obj)
{
	pbs_db_resv_info_t *presv = obj->pbs_db_un.pbs_db_resv;
	PGresult *res;
	int rc;

	SET_PARAM_STR(conn_data, presv->ri_resvid, 0);

	if ((rc = db_query(conn, STMT_SELECT_RESV, 1, &res)) != 0)
		return rc;

	rc = load_resv(res, presv, 0);

	PQclear(res);

	return rc;
}

/**
 * @brief
 *	Find resv
 *
 * @param[in]	conn - Connection handle
 * @param[in]	st   - The cursor state variable updated by this query
 * @param[in]	obj  - Information of resv to be found
 * @param[in]	opts - Any other options (like flags, timestamp)
 *
 * @return      Error code
 * @retval	-1 - Failure
 * @retval	 0 - Success
 * @retval	 1 - Success, but no rows found
 *
 */
int
pbs_db_find_resv(void *conn, void *st, pbs_db_obj_info_t *obj,
		 pbs_db_query_options_t *opts)
{
	PGresult *res;
	int rc;
	int params;
	char conn_sql[MAX_SQL_LENGTH];
	db_query_state_t *state = (db_query_state_t *) st;

	if (!state)
		return -1;

	params = 0;
	strcpy(conn_sql, STMT_FINDRESVS_ORDBY_CREATTM);

	if ((rc = db_query(conn, conn_sql, params, &res)) != 0)
		return rc;

	state->row = 0;
	state->res = res;
	state->count = PQntuples(res);
	return 0;
}

/**
 * @brief
 *	Get the next resv from the cursor
 *
 * @param[in]	conn - Connection handle
 * @param[in]	st   - The cursor state
 * @param[in]	obj  - Resv information is loaded into this object
 *
 * @return      Error code
 *		(Even though this returns only 0 now, keeping it as int
 *			to support future change to return a failure)
 * @retval	 0 - Success
 *
 */
int
pbs_db_next_resv(void *conn, void *st, pbs_db_obj_info_t *obj)
{
	db_query_state_t *state = (db_query_state_t *) st;
	return (load_resv(state->res, obj->pbs_db_un.pbs_db_resv, state->row));
}

/**
 * @brief
 *	Delete the resv from the database
 *
 * @param[in]	conn - Connection handle
 * @param[in]	obj  - Resv information
 *
 * @return      Error code
 * @retval	-1 - Failure
 * @retval	 0 - Success
 *
 */
int
pbs_db_delete_resv(void *conn, pbs_db_obj_info_t *obj)
{
	pbs_db_resv_info_t *presv = obj->pbs_db_un.pbs_db_resv;
	SET_PARAM_STR(conn_data, presv->ri_resvid, 0);
	return (db_cmd(conn, STMT_DELETE_RESV, 1));
}

/**
 * @brief
 *	Deletes attributes of a Resv
 *
 * @param[in]	conn - Connection handle
 * @param[in]	obj_id  - Resv id
 * @param[in]	attr_list - List of attributes
 *
 * @return      Error code
 * @retval	 0 - Success
 * @retval	-1 - On Failure
 *
 */
int
pbs_db_del_attr_resv(void *conn, void *obj_id, pbs_db_attr_list_t *attr_list)
{
	char *raw_array = NULL;
	int len = 0;
	int rc = 0;

	if ((len = attrlist_to_dbarray_ex(&raw_array, attr_list, 1)) <= 0)
		return -1;

	SET_PARAM_STR(conn_data, obj_id, 0);
	SET_PARAM_BIN(conn_data, raw_array, len, 1);

	rc = db_cmd(conn, STMT_REMOVE_RESVATTRS, 2);

	return rc;
}
