/*
 * Copyright (C) 1994-2021 Altair Engineering, Inc.
 * For more information, contact Altair at www.altair.com.
 *
 * This file is part of both the OpenPBS software ("OpenPBS")
 * and the PBS Professional ("PBS Pro") software.
 *
 * Open Source License Information:
 *
 * OpenPBS is free software. You can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 *
 * OpenPBS is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Commercial License Information:
 *
 * PBS Pro is commercially licensed software that shares a common core with
 * the OpenPBS software.  For a copy of the commercial license terms and
 * conditions, go to: (http://www.pbspro.com/agreement.html) or contact the
 * Altair Legal Department.
 *
 * Altair's dual-license business model allows companies, individuals, and
 * organizations to create proprietary derivative works of OpenPBS and
 * distribute them - whether embedded or bundled with other software -
 * under a commercial license agreement.
 *
 * Use of Altair's trademarks, including but not limited to "PBS™",
 * "OpenPBS®", "PBS Professional®", and "PBS Pro™" and Altair's logos is
 * subject to Altair's trademark licensing policies.
 */

/** @file	int_status.c
 * @brief
 * The function that underlies all the status requests
 */

#include <pbs_config.h> /* the master config generated by configure */

#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include "libpbs.h"
#include "pbs_ecl.h"
#include "libutil.h"
#include "attribute.h"
#include "cmds.h"
#include "pbs_internal.h"

/**
 * @brief
 *	-wrapper function for PBSD_status_put which sends
 *	status batch request
 *
 * @param[in] c - socket descriptor
 * @param[in] function - request type
 * @param[in] id - object id
 * @param[in] attrib - pointer to attribute list
 * @param[in] extend - extention string for req encode
 *
 * @return	structure handle
 * @retval 	pointer to batch status on SUCCESS
 * @retval 	NULL on failure
 *
 */
struct batch_status *
PBSD_status(int c, int function, const char *objid, struct attrl *attrib, const char *extend)
{
	int rc;

	/* send the status request */

	if (objid == NULL)
		objid = ""; /* set to null string for encoding */

	rc = PBSD_status_put(c, function, objid, attrib, extend, PROT_TCP, NULL);
	if (rc) {
		return NULL;
	}

	/* get the status reply */
	return PBSD_status_get(c);
}

/**
 * @brief
 *	Returns pointer to status record
 *
 * @param[in] c - connection socket
 * @param[out] last - return pointer to the last batch status read
 *
 * @return returns a pointer to a batch_status structure
 * @retval pointer to batch status on SUCCESS
 * @retval NULL on failure
 */
struct batch_status *
PBSD_status_get(int c)
{
	struct batch_status *rbsp = NULL;
	struct batch_reply *reply;

	/* read reply from stream into presentation element */

	reply = PBSD_rdrpy(c);
	if (reply == NULL) {
		if (pbs_errno == PBSE_NONE)
			pbs_errno = PBSE_PROTOCOL;
		goto end;
	} else if (reply->brp_choice != BATCH_REPLY_CHOICE_NULL &&
		   reply->brp_choice != BATCH_REPLY_CHOICE_Text &&
		   reply->brp_choice != BATCH_REPLY_CHOICE_Status) {
		if (pbs_errno == PBSE_NONE)
			pbs_errno = PBSE_PROTOCOL;
		goto end;
	} else if (get_conn_errno(c) == 0) {
		rbsp = reply->brp_un.brp_statc;
		reply->brp_un.brp_statc = NULL;
	}

end:
	PBSD_FreeReply(reply);
	return rbsp;
}
