/*
 * Copyright (C) 1994-2021 Altair Engineering, Inc.
 * For more information, contact Altair at www.altair.com.
 *
 * This file is part of both the OpenPBS software ("OpenPBS")
 * and the PBS Professional ("PBS Pro") software.
 *
 * Open Source License Information:
 *
 * OpenPBS is free software. You can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 *
 * OpenPBS is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Commercial License Information:
 *
 * PBS Pro is commercially licensed software that shares a common core with
 * the OpenPBS software.  For a copy of the commercial license terms and
 * conditions, go to: (http://www.pbspro.com/agreement.html) or contact the
 * Altair Legal Department.
 *
 * Altair's dual-license business model allows companies, individuals, and
 * organizations to create proprietary derivative works of OpenPBS and
 * distribute them - whether embedded or bundled with other software -
 * under a commercial license agreement.
 *
 * Use of Altair's trademarks, including but not limited to "PBS™",
 * "OpenPBS®", "PBS Professional®", and "PBS Pro™" and Altair's logos is
 * subject to Altair's trademark licensing policies.
 */
#include <pbs_config.h> /* the master config generated by configure */

#include "pbs_nodes.h"
#include "svrfunc.h"
#include "tpp.h"

/**
 * @file	daemon_info.c
 * @brief
 * 		mom_info.c - functions relating to the daemon_info structures
 *
 *		Functions are used for both mom and peer server
 */

#ifndef PBS_MOM /* Not Mom, i.e. the Server */

/**
 * @brief initialize daemon info structure
 * This struct is common for all service end points
 * inlcluding mom/peer-svr
 *
 * @param[in] pul - list of IP addresses of host; will be freed on error
 *			or saved in structure; caller must not free pul
 * @param[in] port - port of service end point
 * @param[in] pmi - machine info struct
 * @return dmn_info_t*
 */
dmn_info_t *
init_daemon_info(unsigned long *pul, uint port, struct machine_info *pmi)
{
	dmn_info_t *dmn_info = calloc(1, sizeof(dmn_info_t));
	if (!dmn_info)
		return NULL;

	dmn_info->dmn_state = INUSE_UNKNOWN | INUSE_DOWN | INUSE_NEEDS_HELLOSVR;
	dmn_info->dmn_stream = -1;
	CLEAR_HEAD(dmn_info->dmn_deferred_cmds);
	dmn_info->dmn_addrs = pul;

	while (*pul) {
		tinsert2(*pul, port, pmi, &ipaddrs);
		pul++;
	}

	return dmn_info;
}

/**
 * @brief free up daemon info struct and associated data
 *
 * @param[in] pmi - mom/peer-svr struct
 */
void
delete_daemon_info(struct machine_info *pmi)
{
	dmn_info_t *pdmninfo;
	unsigned long *up;

	if (!pmi || !pmi->mi_dmn_info)
		return;

	pdmninfo = pmi->mi_dmn_info;

	/* take stream out of tree */
	tpp_close(pdmninfo->dmn_stream);
	tdelete2((unsigned long) pdmninfo->dmn_stream, 0, &streams);

	if (pdmninfo->dmn_addrs) {
		for (up = pdmninfo->dmn_addrs; *up; up++) {
			/* del Mom's IP addresses from tree  */
			tdelete2(*up, pmi->mi_port, &ipaddrs);
		}
		free(pdmninfo->dmn_addrs);
		pdmninfo->dmn_addrs = NULL;
	}

	free(pdmninfo);
	pmi->mi_dmn_info = NULL;
}

#endif /* end of server functions */
